﻿#include "define.h"

// INCLUDE
	vec4 computePixelColor();
// IN
	uniform int bucketFragments;

	layout(binding = 0, r32ui) coherent  uniform uimage2DArray		image_next;
	layout(binding = 1, r32ui) writeonly uniform uimageBuffer		image_pointers;
	layout(binding = 4, rg32f) writeonly uniform  imageBuffer		image_peel;
	layout(binding = 5, offset = 0)		 uniform atomic_uint		next_address[BUCKET_SIZE];
	layout(binding = 10		 )			 uniform sampler2DRect		tex_depth_bounds;

	uint exchangePixelCurrentPageID	(const ivec3 coords, const uint val)	{ return imageAtomicExchange(image_next, coords, val);}
	void sharedPoolSetValue			(const uint index  , const vec4 val)	{		 imageStore(image_peel    , int(index), val);}
	void sharedPoolSetLink			(const uint index  , const uint val)	{		 imageStore(image_pointers, int(index), uvec4(val, 0U, 0U, 0U) );}

void main(void)
{
	ivec2 coords = ivec2(gl_FragCoord.xy);

	// Compute Fragment Value
	vec4  value  = vec4(packUnorm4x8(computePixelColor().abgr), gl_FragCoord.z, 0.0f, 0.0f);

	// Find Bucket 
	vec2	depth_bounds = texture	 (tex_depth_bounds, gl_FragCoord.xy).rg;
	float	depth_near   = -depth_bounds.x;
	float	depth_length =  depth_bounds.y - depth_near;
	int		bucket		 = int((float(BUCKET_SIZE)*((gl_FragCoord.z-depth_near)/depth_length))); if (bucket == BUCKET_SIZE) bucket = BUCKET_SIZE-1;

	// Page ID
	uint page_id, next_id, glob_pID;

	page_id = atomicCounterIncrement(next_address[bucket]);
	next_id = exchangePixelCurrentPageID(ivec3(coords,bucket), page_id);
	
	glob_pID = uint(bucketFragments*bucket);
	page_id += glob_pID;
	next_id	+= glob_pID;

	sharedPoolSetLink  (page_id, next_id);
	sharedPoolSetValue (page_id, value  );
}