﻿#include "define.h"

	vec4 computePixelColor();

	layout(binding = 0, r32ui) coherent uniform uimage2DRect  image_counter;
	layout(binding = 1, r32ui) coherent uniform uimage2DRect  image_semaphore;
	layout(binding = 2, r32f ) coherent uniform  image2DArray image_peel_depth;
	layout(binding = 3, r32ui) coherent uniform uimage2DArray image_peel_color;

	float getPixelFragDepthValue(const int coord_z					 ){return imageLoad	(image_peel_depth, ivec3(gl_FragCoord.xy, coord_z)).r;}
	uint  getPixelFragColorValue(const int coord_z					 ){return imageLoad	(image_peel_color, ivec3(gl_FragCoord.xy, coord_z)).r;}

	void  setPixelFragDepthValue(const int coord_z, const float depth){imageStore (image_peel_depth, ivec3(gl_FragCoord.xy, coord_z),  vec4(depth, 0.0f, 0.0f, 0.0f));}
	void  setPixelFragColorValue(const int coord_z, const uint  color){imageStore (image_peel_color, ivec3(gl_FragCoord.xy, coord_z), uvec4(color, 0U  , 0U  , 0U  ));}

	void  setPixelFragCounter	(const uvec4 val)					 {	      imageStore (image_counter, ivec2(gl_FragCoord.xy), val);}
	uint  getPixelFragCounter	(				)					 { return imageLoad  (image_counter, ivec2(gl_FragCoord.xy)).r;}

	bool  semaphoreAcquire		(								   ) { return (imageLoad (image_semaphore, ivec2(gl_FragCoord.xy)).r == 1U) ? false :
																			   imageAtomicExchange (image_semaphore, ivec2(gl_FragCoord.xy), 1U)==0U;}
	void  semaphoreRelease		(								   ) {		   imageStore(image_semaphore, ivec2(gl_FragCoord.xy), uvec4(0U));}

	vec2 getMaxFromGlobalArray()
	{
		vec2 maxFR = vec2(-1.0f,0.0f);
		
		for(int i=HEAP_SIZE_1n; i>0; i--)
		{
			float Zi = getPixelFragDepthValue(i);
			if(maxFR.g < Zi)
			{
				maxFR.r = i;
				maxFR.g = Zi;
			}
		}	

		return maxFR;
	}

	int setMaxFromGlobalArray()
	{
		int id;
		vec2 maxFR = getMaxFromGlobalArray();

		if(gl_FragCoord.z < maxFR.g)
		{
			id = int(maxFR.r);
			setPixelFragDepthValue(0, maxFR.g);
			setPixelFragColorValue(0, getPixelFragColorValue(id));
		}
		else
			id = 0;

		return id;
	}

	void insertFragmentArray()
	{
		int id = int(getPixelFragCounter());
		if(id < HEAP_SIZE)
		{
			setPixelFragCounter(uvec4(id+1U, 0U, 0U, 0U));
			if(id == HEAP_SIZE_1n)
				id = setMaxFromGlobalArray();
			else
				id = HEAP_SIZE_1n-id;
		}
		else
		{
			if(gl_FragCoord.z < getPixelFragDepthValue(0))
				id = setMaxFromGlobalArray();
			else
				return;
		}

		setPixelFragDepthValue(id, gl_FragCoord.z);
		setPixelFragColorValue(id, packUnorm4x8(computePixelColor()));
	}

	void main(void)
	{
		if(getPixelFragCounter() < HEAP_SIZE || gl_FragCoord.z < getPixelFragDepthValue(0))
		{
			bool stay_loop = true;
			while (stay_loop)
			{
				if (semaphoreAcquire())
				{
					insertFragmentArray();

					semaphoreRelease();
					stay_loop = false;
				}
			}
		}
		discard;
	}