﻿#include "define.h"

// INCLUDE
	vec4 computePixelColor();
// IN
	layout(binding = 0, r32ui) coherent uniform uimage2DRect  image_counter;
	layout(binding = 1, r32ui) coherent uniform uimage2DRect  image_semaphore;
	layout(binding = 2, r32f ) coherent uniform  image2DArray image_peel_depth;
	layout(binding = 3, r32ui) coherent uniform uimage2DArray image_peel_color;

#if multipass
	layout(binding = 4)					uniform  sampler2DRect	 tex_depth;
	float getPixelDepthPrevMax  (									)	{return texture    (tex_depth, gl_FragCoord.xy).r;}
#endif
	void  setPixelFragDepthValue(const int coord_z, const  vec4 val	)	{		imageStore (image_peel_depth, ivec3(gl_FragCoord.xy, coord_z), val);}
	vec4  getPixelFragDepthValue(const int coord_z					)	{return imageLoad  (image_peel_depth, ivec3(gl_FragCoord.xy, coord_z));}

	void  setPixelFragColorValue(const int coord_z, const uvec4 val	)	{		imageStore (image_peel_color, ivec3(gl_FragCoord.xy, coord_z), val);}
	uvec4 getPixelFragColorValue(const int coord_z					)	{return imageLoad  (image_peel_color, ivec3(gl_FragCoord.xy, coord_z));}

	void  setPixelFragCounter	(const uvec4 val					)	{		imageStore (image_counter, ivec2(gl_FragCoord.xy), val);}
	uint  getPixelFragCounter	(									)	{return imageLoad  (image_counter, ivec2(gl_FragCoord.xy)).r;}

	bool  semaphoreAcquire		(									)	{return (imageLoad			(image_semaphore, ivec2(gl_FragCoord.xy)).r == 1U) ? false :
																				 imageAtomicExchange(image_semaphore, ivec2(gl_FragCoord.xy), 1U)==0U;}
	void  semaphoreRelease		(									)	{		 imageStore			(image_semaphore, ivec2(gl_FragCoord.xy), uvec4(0U));}

	layout(location = 0, index = 0) out vec4 out_frag_avg_color;

	int  parentHeap	(const int j) {return (j-1) >> 1;}
	int  leftHeap	(const int j) {return (j<<1) + 1;}

	void maxHeapify()
	{
		if(gl_FragCoord.z >= getPixelFragDepthValue(0).r)
			return;

		int    P;
		int	   I = 0;
		int    L = 1;
		int    R = 2;

		vec4  Zl = getPixelFragDepthValue(L);
		vec4  Zr = getPixelFragDepthValue(R);

		for (int i=0; i<HEAP_SIZE_LOG2; i++)
		{
			float Z = gl_FragCoord.z;
			if(L < HEAP_SIZE && Z < Zl.r)
			{
				P = L;
				Z = Zl.r;
			}
			else
				P = I;

			if(R < HEAP_SIZE && Z < Zr.r)
				P = R;

			if(P != I)
			{
				setPixelFragDepthValue(I, getPixelFragDepthValue(P));
				setPixelFragColorValue(I, getPixelFragColorValue(P));

				I = P; 
				L = leftHeap(I);	Zl = getPixelFragDepthValue(L);
				R = L + 1;			Zr = getPixelFragDepthValue(R);
			}
			else
				break;
		}

		setPixelFragDepthValue(I,  vec4(gl_FragCoord.z, 0.0f, 0.0f, 0.0f));
		setPixelFragColorValue(I, uvec4(packUnorm4x8(computePixelColor()), 0U, 0U, 0U));
	}

	void insertFragmentHeap()
	{
		int counter = int(getPixelFragCounter());

		if (counter == HEAP_SIZE)
			maxHeapify();
		else
		{
			int i = counter;
			int p = parentHeap(i);

			vec4 Zp;
			while (i > 0 && gl_FragCoord.z > (Zp = getPixelFragDepthValue(p)).r)
			{
				setPixelFragDepthValue(i, Zp);
				setPixelFragColorValue(i, getPixelFragColorValue(p));

				i = p;
				p = parentHeap(i);
			}

			setPixelFragDepthValue(i,  vec4(gl_FragCoord.z, 0.0f, 0.0f, 0.0f));
			setPixelFragColorValue(i, uvec4(packUnorm4x8(computePixelColor()), 0U, 0U, 0U));

			setPixelFragCounter(uvec4(uint(counter)+1U, 0U, 0U, 0U));
		}
	}

	void main(void)
	{

#if multipass
	if(gl_FragCoord.z <= getPixelDepthPrevMax())
		discard;
#endif

		if(getPixelFragCounter() < HEAP_SIZE || gl_FragCoord.z < getPixelFragDepthValue(0).r)
		{
			bool stay_loop = true;
			while (stay_loop)
			{
				if (semaphoreAcquire())
				{
					insertFragmentHeap();

					semaphoreRelease();
					stay_loop = false;
				}

			}
		}

		discard;
	}