﻿#include "define.h"
#include "s-buffer.h"

	vec4 computePixelColor();

	uniform uint *final_address[COUNTERS];

	layout(binding = 0, r32ui) coherent uniform uimage2DRect  image_counter;
	layout(binding = 1, r32ui) readonly uniform uimage2DRect  image_next;
	layout(binding = 2, r32f ) coherent uniform  imageBuffer  image_peel_depth;
	layout(binding = 3, r32ui) coherent uniform uimageBuffer  image_peel_color;
	layout(binding = 4, r32ui) coherent uniform uimage2DRect  image_semaphore;
	layout(binding = 5, r32ui) readonly uniform uimage2DRect  image_counter_total;

	uint getPixelFragNextAddress	(								  ) { return imageLoad  (image_next   , ivec2(gl_FragCoord.xy)).r;}
	uint getPixelFragCounter		(								  ) { return imageLoad  (image_counter, ivec2(gl_FragCoord.xy)).r;}
	uint getPixelFragCounterTotal	(								  ) { return imageLoad  (image_counter_total, ivec2(gl_FragCoord.xy)).r;}
	void setPixelFragCounter		(const uvec4 val				  ) {		 imageStore (image_counter, ivec2(gl_FragCoord.xy), val);}

	float sharedPoolGetDepthValue	(const uint index				  ) { return imageLoad	(image_peel_depth, int(index)).r;}
	uint  sharedPoolGetColorValue	(const uint index				  ) { return imageLoad	(image_peel_color, int(index)).r;}
	void  sharedPoolSetDepthValue	(const uint index, const float val) {		 imageStore	(image_peel_depth, int(index), vec4 (val,0.0f,0.0f,0.0f));}
	void  sharedPoolSetColorValue	(const uint index, const uint  val) {		 imageStore	(image_peel_color, int(index), uvec4(val,  0U,  0U,  0U));}

	bool semaphoreAcquire			(								  ) {return (imageLoad			 (image_semaphore, ivec2(gl_FragCoord.xy)).r == 1U) ? false :
																				 imageAtomicExchange (image_semaphore, ivec2(gl_FragCoord.xy), 1U)==0U;}
	void semaphoreRelease			(								  ) {		 imageStore			 (image_semaphore, ivec2(gl_FragCoord.xy), uvec4(0U));}

	vec2 getMaxFromGlobalArray(uint start)
	{
		vec2 maxFR = vec2(-1.0f, 0.0f);
		
		for(uint s=start+HEAP_SIZE_1n; s>start; s--)
		{
			float Zi = sharedPoolGetDepthValue(s);
			if(maxFR.g < Zi)
			{
				maxFR.r = s;
				maxFR.g = Zi;
			}
		}	

		return maxFR;
	}

	uint setMaxFromGlobalArray(uint start)
	{	
		uint id;
		vec2 maxFR = getMaxFromGlobalArray(start);

		if(gl_FragCoord.z < maxFR.g)
		{
			id = uint(maxFR.r);
			sharedPoolSetDepthValue(start, maxFR.g);
			sharedPoolSetColorValue(start, sharedPoolGetColorValue(id));
		}
		else
			id  = start;

		return id;
	}

	void insertFragmentArray(uint start)
	{
		uint  id, counter = getPixelFragCounter();

		if(counter < HEAP_SIZE)
		{	
			if(counter == HEAP_SIZE_1n)
			{
				sharedPoolSetDepthValue(start+HEAP_SIZE_1n, sharedPoolGetDepthValue(start));
				sharedPoolSetColorValue(start+HEAP_SIZE_1n, sharedPoolGetColorValue(start));
				sharedPoolSetDepthValue(start, 1.0f);

				id = setMaxFromGlobalArray(start);
			}
			else
				id = start + counter;
			setPixelFragCounter(uvec4(counter+1U, 0U, 0U, 0U));
		}
		else
		{
			if(gl_FragCoord.z < sharedPoolGetDepthValue(start))
				id = setMaxFromGlobalArray(start);
			else
				return;
		}

		sharedPoolSetDepthValue(id, gl_FragCoord.z);
		sharedPoolSetColorValue(id, packUnorm4x8(computePixelColor()));
	}

	uint  parentHeap(const uint j) {return (j-1U) >> 1U;}
	uint  leftHeap	(const uint j) {return (j<<1U) + 1U;}

	void maxHeapify(uint start)
	{
		if(gl_FragCoord.z >= sharedPoolGetDepthValue(start).r)
			return;

		uint    P;
		uint	I = 0U;
		uint    L = 1U;
		uint    R = 2U;

		uint	idI = start;
		uint	idP;

		float	Zl = sharedPoolGetDepthValue(start + L);
		float	Zr = sharedPoolGetDepthValue(start + R);

		for (int j=0; j<HEAP_SIZE_LOG2; j++)
		{
			float Z = gl_FragCoord.z;
			if(L < HEAP_SIZE && Z < Zl)
			{
				P = L;
				Z = Zl;
			}
			else
				P = I;

			if(R < HEAP_SIZE && Z < Zr)
				P = R;

			if(P != I)
			{
				idP = start + P;

				sharedPoolSetDepthValue(idI, sharedPoolGetDepthValue(idP));
				sharedPoolSetColorValue(idI, sharedPoolGetColorValue(idP));

				I = P;				idI = start + I;
				L = leftHeap(I);	Zl  = sharedPoolGetDepthValue(start + L);
				R = L + 1U;			Zr  = sharedPoolGetDepthValue(start + R);
			}
			else
				break;
		}

		sharedPoolSetDepthValue(idI, gl_FragCoord.z);
		sharedPoolSetColorValue(idI, packUnorm4x8(computePixelColor()));
	}

	void insertFragmentHeap(uint start)
	{
		uint counter = getPixelFragCounter();

		if (counter == HEAP_SIZE)
			maxHeapify(start);
		else
		{
			uint i = counter;
			uint p = parentHeap(i);

			uint  id;
			float Zp;
			while (i > 0U && gl_FragCoord.z > (Zp = sharedPoolGetDepthValue(start + p)))
			{
				id = start + i;
				sharedPoolSetDepthValue(id, Zp);
				sharedPoolSetColorValue(id, sharedPoolGetColorValue(start + p));

				i = p;
				p = parentHeap(i);
			}
			
			id = start + i;
			sharedPoolSetDepthValue(id, gl_FragCoord.z);
			sharedPoolSetColorValue(id, packUnorm4x8(computePixelColor()));

			setPixelFragCounter(uvec4(counter+1U, 0U, 0U, 0U));
		}
	}

	void main(void)
	{
		uint  address = getPixelFragNextAddress();
		int   hash_id = hashFunction(ivec2(gl_FragCoord.xy));
		uint  start   = (*final_address[hash_id]) + address;

		uint counter_total = getPixelFragCounterTotal();

		if(counter_total < HEAP_SIZE || getPixelFragCounter() < HEAP_SIZE || gl_FragCoord.z < sharedPoolGetDepthValue(start))
		{
			bool stay_loop = true;
			while (stay_loop)
			{
				if (semaphoreAcquire())
				{
					if(counter_total < HEAP_SIZE || HEAP_SIZE < 16)
						insertFragmentArray(start);
					else
						insertFragmentHeap(start);

					semaphoreRelease();
					stay_loop = false;
				}
			}
		}
		discard;
	}