﻿#include "define.h"
#include "s-buffer.h"
#include "sort.h"
#include "resolve.h"

	uniform uint  total_fragments;
	uniform uint *final_address[COUNTERS];

#if atomic
	layout(binding = 0, r32ui) readonly  uniform uimage2DRect  image_counter;
#else
	layout(binding = 0)					 uniform sampler2DRect tex_counter;
#endif
	layout(binding = 1, r32ui) readonly  uniform uimage2DRect  image_next;
	layout(binding = 2, r32f ) readonly  uniform  imageBuffer  image_peel_depth;
	layout(binding = 3, r32ui) readonly  uniform uimageBuffer  image_peel_color;

	float sharedPoolGetDepthValue	(const uint index) {return imageLoad (image_peel_depth, int(index)).x;}
	uint  sharedPoolGetColorValue	(const uint index) {return imageLoad (image_peel_color, int(index)).x;}

	uint getPixelNextAddress		(				 ) {return imageLoad(image_next   , ivec2(gl_FragCoord.xy)).x-1U;}
#if atomic
	uint getPixelFragCounter		(				 ) {return imageLoad(image_counter, ivec2(gl_FragCoord.xy)).x;   }
#else
	uint getPixelFragCounter		(				 ) {return uint(texture(tex_counter, ivec2(gl_FragCoord.xy)).x); }
#endif

	layout(location = 0, index = 0) out vec4 out_frag_color;

	void main(void)
	{
		int counter = int(getPixelFragCounter());
		if(counter > 0)
		{	
			uint address = getPixelNextAddress();
			int  hash_id = hashFunction(ivec2(gl_FragCoord.xy));
#if inverse	
			bool front_prefix = hash_id < COUNTERS_2d ? true : false;
#else
			bool front_prefix = true;
#endif
			uint sum = (*final_address[hash_id]);

			uint page_id   = front_prefix ? address + sum : total_fragments - address - sum;
			uint direction = front_prefix ? -1U : 1U;
			
			if (counter > HEAP_SIZE)
			{		
				vec3 maxFR = vec3(-1.0f,0.0f,0.0f);
				for(int i=0; i<HEAP_SIZE_1n; i++)
				{
					fragments[i] = vec2(page_id, sharedPoolGetDepthValue(page_id));
					page_id	  += direction;
					if(maxFR.g < fragments[i].g)
					{
						maxFR.r = i;
						maxFR.g = fragments[i].g;
						maxFR.b = fragments[i].r;
					}
				}
				
				vec2 newFR = vec2(page_id, sharedPoolGetDepthValue(page_id));
				page_id	  += direction;
				if(newFR.g < maxFR.g)
				{
					fragments[HEAP_SIZE_1n] = maxFR.bg;
					fragments[int(maxFR.r)] = newFR;
				}
				else
					fragments[HEAP_SIZE_1n] = newFR;

				for(int i=HEAP_SIZE; i<counter; i++)
				{
					newFR	= vec2(page_id, sharedPoolGetDepthValue(page_id));
					page_id+= direction;
					if(newFR.g < fragments[HEAP_SIZE_1n].g)
					{
						maxFR = getMaxFromLocalArray3();
						if(newFR.g < maxFR.g)
						{
							fragments[HEAP_SIZE_1n] = maxFR.bg;
							fragments[int(maxFR.r)] = newFR;
						}
						else
							fragments[HEAP_SIZE_1n] = newFR;
					}
				}

				for(int i=0; i<HEAP_SIZE; i++)
				{
					colors	 [i]   = sharedPoolGetColorValue (uint(fragments[i].r));
					fragments[i].r = float(i);
				}
				counter = HEAP_SIZE;
			}
			else
			{
				for(int i=0; i<counter; i++)
				{
					fragments[i] = vec2(i, sharedPoolGetDepthValue(page_id));
					colors	 [i] = sharedPoolGetColorValue(page_id);
					page_id		+= direction;
				}
			}
			out_frag_color = resolve(counter, false);
		}
		else
			discard;
	}