﻿#include "define.h"

	vec4 computePixelColor();

	uniform int  width_2; // 2*screen_width
	uniform uint total_fragments;

	layout(binding = 0, offset = 0)		 uniform atomic_uint  next_address;
	layout(binding = 1, r32ui) coherent  uniform uimageBuffer image_pointers;
	layout(binding = 2, r32f ) coherent	 uniform  imageBuffer image_peel_depth;
	layout(binding = 3, r32ui) writeonly uniform uimageBuffer image_peel_color;	

	float sharedPoolGetDepthValue	(const uint index					)	{return imageLoad (image_peel_depth , int(index)).r;}
	void  sharedPoolSetDepthValue	(const uint index, const float val	)	{		imageStore(image_peel_depth , int(index),  vec4(val, 0.0f, 0.0f, 0.0f));}
	void  sharedPoolSetColorValue	(const uint index, const uint  val	)	{		imageStore(image_peel_color	, int(index), uvec4(val,   0U,   0U,   0U));}
	uint  sharedPoolGetLink			(const uint index					)	{return imageLoad (image_pointers   , int(index)).r;}
	void  sharedPoolSetLink			(const uint index, const uint  val	)	{		imageStore(image_pointers	, int(index), uvec4(val,   0U,   0U,   0U));}
	uint  sharedPoolCompSwapLink	(const uint index, 
									 const uint compare, const uint data)	{return imageAtomicCompSwap(image_pointers, int(index), compare, data);}

	void main(void)
	{
		// New Position
		uint newID = atomicCounterIncrement(next_address);
		if(newID < total_fragments)
		{
			// Prev Position
			uint headID = uint(gl_FragCoord.x + width_2*gl_FragCoord.y) * 2U + 2U;
			uint prevID = headID;

			// Store Color-Depth
			uint  C	= packUnorm4x8(computePixelColor());
			float Z = gl_FragCoord.z;

			int	 iter=0;
			bool stay_loop = true;
			while(stay_loop)
			{
				// Next Position
				uint nextID	= sharedPoolGetLink(prevID);
				if(Z < sharedPoolGetDepthValue(nextID))
				{	
					// 1. [NEW]  --> [NEXT]
					sharedPoolSetLink  (newID, nextID);	
					sharedPoolSetDepthValue(newID, Z);
					sharedPoolSetColorValue(newID, C);

					// 2. [PREV] --> [NEW]
					if(nextID == sharedPoolCompSwapLink(prevID, nextID, newID))
						discard; // success !!!
				}
				else
				{
					if(++iter == HEAP_SIZE)
						discard;
					if(nextID == 0U)
						stay_loop = false;
			
					prevID = nextID;
				}
			}
			discard;
		}
	}